from langchain.chains import LLMRequestsChain, LLMChain
from langchain.prompts import PromptTemplate
from lmchain.lmchain.agents import llmAgent
import requests
from bs4 import BeautifulSoup
from tqdm import tqdm
llm = llmAgent.AgentChatGLM()

class LMRequestsChain:
    def __init__(self,llm,max_url_num = 2):
        template = """Between >>> and <<< are the raw search result text from google.
                Extract the answer to the question '{query}' or say "not found" if the information is not contained.
                Use the format
                Extracted:<answer or "not found">
                >>> {requests_result} <<<
                Extracted:"""
        PROMPT = PromptTemplate(
            input_variables=["query", "requests_result"],
            template=template,
        )
        self.chain = LLMRequestsChain(llm_chain=LLMChain(llm=llm, prompt=PROMPT))
        self.max_url_num = max_url_num

        query_prompt = PromptTemplate(
            input_variables=["query","responses"],
            template = "作为一名专业的信息总结员，我需要查询的信息为{query},根据提供的信息{responses}回答一下查询的结果。")
        self.query_chain = LLMChain(llm=llm, prompt=query_prompt)

    def __call__(self, query,target_site = ""):
        url_list = self.get_urls(query,target_site = target_site)
        print(f"查找到{len(url_list)}条url内容，现在开始解析其中的{self.max_url_num}条内容。")
        responses = []
        for url in tqdm(url_list[:self.max_url_num]):
            inputs = {
                "query": query,
                "url": url
            }

            response = self.chain(inputs)
            output = response["output"]
            responses.append(output)
        if len(responses) != 0:
            output = self.query_chain.run({"query":query,"responses":responses})
            return output
        else:
            return "查找内容为空，请更换查找词"

    def query_form_url(self,query = "LMchain是什么？",url = ""):
        assert url != "",print("url link must be set")
        inputs = {
            "query": query,
            "url": url
        }
        response = self.chain(inputs)
        return response

    def get_urls(self,query='lmchain是什么?', target_site=""):
        def bing_search(query, count=30):
            url = f'https://cn.bing.com/search?q={query}'
            headers = {
                'User-Agent': 'Mozilla/6.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/58.0.3029.110 Safari/537.3'}
            response = requests.get(url, headers=headers)
            if response.status_code == 200:
                html = response.text
                # 使用BeautifulSoup解析HTML

                soup = BeautifulSoup(html, 'html.parser')
                results = soup.find_all('li', class_='b_algo')
                return [result.find('a').text for result in results[:count]]
            else:
                print(f'请求失败，状态码：{response.status_code}')
                return []
        results = bing_search(query)
        if len(results) == 0:
            return None
        url_list = []
        if target_site != "":
            for i, result in enumerate(results):
                if "https" in result and target_site in result:
                    url = "https://" + result.split("https://")[1]
                    url_list.append(url)
        else:
            for i, result in enumerate(results):
                if "https" in result:
                    url = "https://" + result.split("https://")[1]
                    url_list.append(url)
        if len(url_list) > 0:
            return url_list
        else:
            # 这里是确保在知乎里面找不到对应的内容，有相应的内容返回
            for i, result in enumerate(results):
                if "https" in result:
                    url = "https://" + result.split("https://")[1]
                    url_list.append(url)
            return url_list

if __name__ == '__main__':
    llm = llmAgent.AgentChatGLM()
    response = LMRequestsChain(llm=llm)("酶切位点")
    print(response)




import asyncio
from abc import ABC, abstractmethod
from typing import List


class Embeddings(ABC):
    """Interface for embedding models."""

    @abstractmethod
    def embed_documents(self, texts: List[str]) -> List[List[float]]:
        """Embed search docs."""

    @abstractmethod
    def embed_query(self, text: str) -> List[float]:
        """Embed query text."""

    async def aembed_documents(self, texts: List[str]) -> List[List[float]]:
        """Asynchronous Embed search docs."""
        return await asyncio.get_running_loop().run_in_executor(
            None, self.embed_documents, texts
        )

    async def aembed_query(self, text: str) -> List[float]:
        """Asynchronous Embed query text."""
        return await asyncio.get_running_loop().run_in_executor(
            None, self.embed_query, text
        )

class LMEmbedding(Embeddings):
    from modelscope.pipelines import pipeline
    from modelscope.utils.constant import Tasks
    pipeline_se = pipeline(Tasks.sentence_embedding,model='thomas/text2vec-base-chinese', model_revision='v1.0.0',device="cuda")


    def _costruct_inputs(self,texts):

        inputs = {
                "source_sentence": texts
            }

        return inputs

    def embed_documents(self, texts: List[str]) -> List[List[float]]:
        """Embed search docs."""

        inputs = self._costruct_inputs(texts)
        result_embeddings = self.pipeline_se(input=inputs)
        return result_embeddings["text_embedding"]

    def embed_query(self, text: str) -> List[float]:
        """Embed query text."""
        inputs = self._costruct_inputs([text])
        result_embeddings = self.pipeline_se(input=inputs)
        return result_embeddings["text_embedding"]


class GLMEmbedding(Embeddings):
    import zhipuai as zhipuai
    zhipuai.api_key = "1f565e40af1198e11ff1fd8a5b42771d.SjNfezc40YFsz2KC"  # 控制台中获取的 APIKey 信息
    def _costruct_inputs(self, texts):
        inputs = {
            "source_sentence": texts
        }

        return inputs

    def embed_documents(self, texts: List[str]) -> List[List[float]]:
        """Embed search docs."""
        result_embeddings = []
        for text in texts:
            embedding = self.embed_query(text)
            result_embeddings.append(embedding)
        return result_embeddings

    def embed_query(self, text: str) -> List[float]:
        """Embed query text."""
        result_embeddings = 	self.zhipuai.model_api.invoke(
            model="text_embedding",        prompt=text)
        return result_embeddings["data"]["embedding"]




if __name__ == '__main__':
    inputs = ["不可以，早晨喝牛奶不科学","不可以，今天早晨喝牛奶不科学","早晨喝牛奶不科学"]
    print(GLMEmbedding().embed_documents(inputs))





import inspect
import traceback
from copy import deepcopy
from pprint import pformat
from types import GenericAlias
from typing import get_origin, Annotated

_TOOL_HOOKS = {}
_TOOL_DESCRIPTIONS = {}


def register_tool(func: callable):
    tool_name = func.__name__
    tool_description = inspect.getdoc(func).strip()
    python_params = inspect.signature(func).parameters
    tool_params = []
    for name, param in python_params.items():
        annotation = param.annotation
        if annotation is inspect.Parameter.empty:
            raise TypeError(f"Parameter `{name}` missing type annotation")
        if get_origin(annotation) != Annotated:
            raise TypeError(f"Annotation type for `{name}` must be typing.Annotated")

        typ, (description, required) = annotation.__origin__, annotation.__metadata__
        typ: str = str(typ) if isinstance(typ, GenericAlias) else typ.__name__
        if not isinstance(description, str):
            raise TypeError(f"Description for `{name}` must be a string")
        if not isinstance(required, bool):
            raise TypeError(f"Required for `{name}` must be a bool")

        tool_params.append({
            "name": name,
            "description": description,
            "type": typ,
            "required": required
        })
    tool_def = {
        "name": tool_name,
        "description": tool_description,
        "params": tool_params
    }

    # print("[registered tool] " + pformat(tool_def))
    _TOOL_HOOKS[tool_name] = func
    _TOOL_DESCRIPTIONS[tool_name] = tool_def

    return func


def dispatch_tool(tool_name: str, tool_params: dict) -> str:
    if tool_name not in _TOOL_HOOKS:
        return f"Tool `{tool_name}` not found. Please use a provided tool."
    tool_call = _TOOL_HOOKS[tool_name]
    try:
        ret = tool_call(**tool_params)
    except:
        ret = traceback.format_exc()
    return str(ret)


def get_tools() -> dict:
    return deepcopy(_TOOL_DESCRIPTIONS)


# Tool Definitions

# @register_tool
# def random_number_generator(
#         seed: Annotated[int, 'The random seed used by the generator', True],
#         range: Annotated[tuple[int, int], 'The range of the generated numbers', True],
# ) -> int:
#     """
#     Generates a random number x, s.t. range[0] <= x < range[1]
#     """
#     if not isinstance(seed, int):
#         raise TypeError("Seed must be an integer")
#     if not isinstance(range, tuple):
#         raise TypeError("Range must be a tuple")
#     if not isinstance(range[0], int) or not isinstance(range[1], int):
#         raise TypeError("Range must be a tuple of integers")
#
#     import random
#     return random.Random(seed).randint(*range)
#
#
# @register_tool
# def get_weather(
#         city_name: Annotated[str, 'The name of the city to be queried', True],
# ) -> str:
#     """
#     Get the current weather for `city_name`
#     """
#
#     if not isinstance(city_name, str):
#         raise TypeError("City name must be a string")
#
#     key_selection = {
#         "current_condition": ["temp_C", "FeelsLikeC", "humidity", "weatherDesc", "observation_time"],
#     }
#     import requests
#     try:
#         resp = requests.get(f"https://wttr.in/{city_name}?format=j1")
#         resp.raise_for_status()
#         resp = resp.json()
#         ret = {k: {_v: resp[k][0][_v] for _v in v} for k, v in key_selection.items()}
#     except:
#         import traceback
#         ret = "Error encountered while fetching weather data!\n" + traceback.format_exc()
#
#     return str(ret)
#
#
# @register_tool
# def get_customer_weather(location: Annotated[str, "需要查询位置的名称，用中文表示的地点名称", True] = ""):
#     """ 自己编写的天气查询函数"""
#
#     if location == "上海":
#         return 23.0
#     elif location == "南京":
#         return 25.0
#     else:
#         return "未查询相关内容"
#
#
# @register_tool
# def get_random_fun(location: Annotated[str, "随机参数", True] = ""):
#     """编写的一个混淆随机函数"""
#     location = location
#     return "你上当啦"


if __name__ == "__main__":
    print(dispatch_tool("get_weather", {"city_name": "shanghai"}))
    print(get_tools())



# LM AI Similarity Search
import torch
import math


class LMASS:
    def __init__(self):
        super().__init__()
        self.texts = []
        self.vectors = []
        self.embedding_class = None

    def from_documents(self,docs,embedding_class):

        texts =  [doc.page_content for doc in docs]
        self.texts = texts
        self.docs = docs
        self.embedding_class = embedding_class

        self.vectors = embedding_class.embed_documents(texts)
        return self.vectors

    def get_relevant_documents(self,query,k = 3):
        query_vector = self.embedding_class.embed_query(query)
        sorted_index = self.get_similarity_vector_indexs(query_vector,self.vectors,k = k)
        sorted_docs = [self.docs[id] for id in sorted_index]

        return sorted_docs


    def get_similarity_vector_indexs(self, query_vector ,vectors, k: int = 3,  ):
        similarity = self._cosine_similaritys(query_vector,vectors)

        #这里是按分值从大到小的进行排序
        # 使用sort()函数对tensor进行降序排序，并返回排序后的tensor和索引
        sorted_tensor, sorted_indices = torch.sort(similarity, descending=True)
        return sorted_indices[:k].numpy()


    def _cosine_similaritys(self, query_vector, vectors):
        query_vector = torch.tensor(query_vector)
        vectors = torch.tensor(vectors)
        similarity_matrix = torch.nn.functional.cosine_similarity(query_vector, vectors, dim=-1)
        return similarity_matrix

        #CosineSimilarity
    def cosine_similarity(self, query_vector, tensor_2):
        normalized_tensor_1 = query_vector / query_vector.norm(dim=-1, keepdim=True)
        normalized_tensor_2 = tensor_2 / tensor_2.norm(dim=-1, keepdim=True)
        return (normalized_tensor_1 * normalized_tensor_2).sum(dim=-1)

    #DotProductSimilarity
    def dot_product_similarity(self, query_vector, tensor_2, scale_output = True):
        result = (query_vector * tensor_2).sum(dim=-1)
        if scale_output:
            # TODO why allennlp do multiplication at here ?
            result /= math.sqrt(query_vector.size(-1))
        return result


if __name__ == '__main__':
    v1 = torch.randn(size=(1,768))
    v2 = torch.randn(size=(4,768))
    lmass = LMASS()
    #print(lmass.dot_product_similarity(v1, v2))
    #print(lmass.cosine_similarity(v1, v2))
    sim_index = lmass.get_similarity_vector_indexs(v1,v2)
    print(sim_index)




from typing import Any, Dict, List, Optional, Type

from langchain.chains.qa_with_sources.retrieval import RetrievalQAWithSourcesChain
from langchain.chains.retrieval_qa.base import RetrievalQA
from langchain.document_loaders.base import BaseLoader
from langchain.pydantic_v1 import BaseModel, Extra, Field
from langchain.schema import Document
from langchain.schema.embeddings import Embeddings
from langchain.schema.language_model import BaseLanguageModel
from langchain.schema.vectorstore import VectorStore
from langchain.text_splitter import RecursiveCharacterTextSplitter, TextSplitter
from langchain.vectorstores.chroma import Chroma


def _get_default_text_splitter() -> TextSplitter:
    return RecursiveCharacterTextSplitter(chunk_size=1000, chunk_overlap=0)

from lmchain.embeddings import embeddings
embedding_tool = embeddings.GLMEmbedding()

class VectorstoreIndexCreator(BaseModel):
    """Logic for creating indexes."""

    class Config:
        """Configuration for this pydantic object."""
        extra = Extra.forbid
        arbitrary_types_allowed = True




    chunk_size = 1280  # 每段字数长度
    chunk_overlap = 32  # 重叠的字数
    text_splitter = RecursiveCharacterTextSplitter(chunk_size=chunk_size, chunk_overlap=chunk_overlap)





    def from_loaders(self, loaders: List[BaseLoader]):
        """Create a vectorstore index from loaders."""
        docs = []
        for loader in loaders:
            docs.extend(loader.load())
        return self.from_documents(docs)


    def from_documents(self, documents: List[Document]):
        #说一下这个index的作用就是返回
        sub_docs = self.text_splitter.split_documents(documents)

        # texts = [d.page_content for d in sub_docs]
        # metadatas = [d.metadata for d in sub_docs]

        qa_chain = ConversationalRetrievalChain(document=sub_docs)
        return qa_chain


from langchain.docstore.document import Document
from langchain.text_splitter import RecursiveCharacterTextSplitter
from lmchain.embeddings import embeddings
from lmchain.vectorstores import laiss
from lmchain.agents import llmMultiAgent
from langchain.memory import ConversationBufferMemory
from langchain.prompts import (
    ChatPromptTemplate,  # 用于构建聊天模板的类
    MessagesPlaceholder,  # 用于在模板中插入消息占位的类
    SystemMessagePromptTemplate,  # 用于构建系统消息模板的类
    HumanMessagePromptTemplate  # 用于构建人类消息模板的类
)
from langchain.chains import ConversationChain

class ConversationalRetrievalChain:
    def __init__(self,document,chunk_size = 1280,chunk_overlap = 50,file_name = "这是一份辅助材料"):
        """
        :param document: 输入的文本内容，只要一个text文本
        :param chunk_size: 切分后每段的字数
        :param chunk_overlap: 每个相隔段落重叠的字数
        :param file_name: 文本名称/文本地址
        """
        self.text_splitter = RecursiveCharacterTextSplitter(    chunk_size=chunk_size, chunk_overlap=chunk_overlap)
        self.embedding_tool = embedding_tool

        self.lmaiss = laiss.LMASS() #这里是用于将文本转化为vector，并且计算query相应的相似度的类
        self.llm = llmMultiAgent.AgentZhipuAI()
        self.memory = ConversationBufferMemory(return_messages=True)

        conversation_prompt = ChatPromptTemplate.from_messages([
                SystemMessagePromptTemplate.from_template("你是一个最强大的人工智能程序，可以知无不答，但是你不懂的东西会直接回答不知道。"),
                MessagesPlaceholder(variable_name="history"),  # 历史消息占位符
                HumanMessagePromptTemplate.from_template("{input}")  # 人类消息输入模板
            ])

        self.qa_chain = ConversationChain(memory=self.memory, prompt=conversation_prompt, llm=self.llm)
        "---------------------------"
        self.metadatas = []
        for doc in document:
            self.metadatas.append(doc.metadata)
        self.documents = self.text_splitter.split_documents(document)    #根据
        self.vectorstore = self.lmaiss.from_documents(self.documents, embedding_class=self.embedding_tool)



    def __call__(self, query):
        query_embedding = self.embedding_tool.embed_query(query)

        #根据query查找最近的那个序列
        close_id = self.lmaiss.get_similarity_vector_indexs(query_embedding, self.vectorstore, k=1)[0]
        #查找最近的那个段落id
        doc = self.documents[close_id]
        meta = self.metadatas[close_id]
        #构建查询的query
        query = f"你现在要回答问题'{query}',你可以参考文献'{doc}',你如果找不到对应的内容,就从自己的记忆体中查找，就回答'请提供更为准确的查询内容'。"
        result = (self.qa_chain.predict(input=query))
        return result,meta


    def query(self,input):
        result,meta = self.__call__(input)
        return result

    #这里的模型的意思是
    def query_with_sources(self,input):
        result,meta = self.__call__(input)
        return {"answer":result,"sources":meta}
